#include "board.h"
#include "qspi.h"
#include "io.h"
#include "common.h"
#include "types.h"
#include "print.h"

static void qspi_supply_clock(void)
{
        u32 sr, cr;

        sr = read32(CPG_MSTPSR9);
        if (sr & CPG_MSTP9_QSPI) {
                cr = read32(CPG_SMSTPCR9);
                write32(CPG_SMSTPCR9, cr & ~CPG_MSTP9_QSPI);
        }
}

static void qspi_pinmux(void)
{
	u32 reg;

	reg = read32(PFC_IPSR2);
	reg &= ~0x0007fff8;
	reg |=  0x000244c8;
	/* SPCLK, MOSI_IO0, MISO_IO1, IO2, IO3, SSL */
	write32(PFC_PMMR, ~reg);
	write32(PFC_IPSR2, reg);

	reg = read32(PFC_GPSR1);
	reg |=  0x000003f0;
	/* GP1[9:4] -> Peripheral(QSPI) */
	write32(PFC_PMMR, ~reg);
	write32(PFC_GPSR1, reg);
}

void qspi_setup(void)
{
	qspi_supply_clock();
	qspi_pinmux();

	/* SPI Function Dsiable */
	write8(QSPI_SPCR, QSPI_SPCR_MSTR);

	write8(QSPI_SSLP, 0x00);
	write8(QSPI_SPPCR, QSPI_SPPCR_IO3FV | QSPI_SPPCR_IO2FV);

	/* Bit Rate: 48.75Mbps*/
	write8(QSPI_SPBR, 0x01);

	write8(QSPI_SPDCR, 0x00);

	/* Clock Delay: 1.5 SPCLK cycles(SPCMD.SCKDEN == 1) */
	write8(QSPI_SPCKD, 0x00);
	/* SSL Negation Delay: 1 SPCLK cycles(SPCMD.SLNDEN == 1) */
	write8(QSPI_SSLND, 0x00);
	/* Next-Access Delay: 1 SPCLK cycles(SPCMD.SPNDEN == 1) */
	write8(QSPI_SPND, 0x00);
}

void qspi_cleanup(void)
{
	/* SPI Function Dsiable */
	write8(QSPI_SPCR, QSPI_SPCR_MSTR);
}

static void qspi_setup_xfer(int len, enum qspi_xfer_state state)
{
	if (state == QSPI_XFER_BEGIN) {
		/* SPI Function Dsiable */
		write8(QSPI_SPCR, QSPI_SPCR_MSTR);

		/* Sequence: 0->0->... */
		write8(QSPI_SPSCR, 0x00);

		/* command */
		write16(QSPI_SPCMD0, QSPI_SPCMD_SCKDEN | QSPI_SPCMD_SLNDEN |
			QSPI_SPCMD_SPNDEN | QSPI_SPCMD_BRDV_DIV2 |
			QSPI_SPCMD_SSLKP);

		/* Transmit/Receive Data Reset */
		write8(QSPI_SPBFCR, QSPI_SPBFCR_TXRST | QSPI_SPBFCR_RXRST);
		write8(QSPI_SPBFCR, 0x00);
	}

	/* transfer size */
	write32(QSPI_SPBMUL0, len);

	/* SPI Function Enable */
	write8(QSPI_SPCR, QSPI_SPCR_SPE | QSPI_SPCR_MSTR);
}

static inline int qspi_is_tx_empty(void)
{
	return read8(QSPI_SPSR) & QSPI_SPSR_SPTEF;
}

static inline int qspi_is_rx_full(void)
{
	return read8(QSPI_SPSR) & QSPI_SPSR_SPRFF;
}

static void qspi_write8(const u8 data)
{
	while (!qspi_is_tx_empty());
	write8(QSPI_SPDR, data);

	while (!qspi_is_rx_full());
	read8(QSPI_SPDR);
}

/* static u8 qspi_read8(const u8 data) */
/* { */
/* 	while (!qspi_is_tx_empty()); */
/* 	write8(QSPI_SPDR, data); */

/* 	while (!qspi_is_rx_full()); */
/* 	return read8(QSPI_SPDR); */
/* } */

void qspi_read_fast(u32 addr, u8 *buf, u32 len)
{
	u32 remain, n, i;
	u8 bfcr;

	qspi_setup_xfer(5, QSPI_XFER_BEGIN);

	/* command */
	qspi_write8(CMD_FAST_READ);
	/* address(24bit) */
	qspi_write8((addr >> 16) & 0xff);
	qspi_write8((addr >>  8) & 0xff);
	qspi_write8( addr        & 0xff);
	/* dummy */
	qspi_write8(DUMMY);

	qspi_setup_xfer(len, QSPI_XFER_END);

	remain = len;
	while (remain) {
		/* Wait transmit */
		while (!qspi_is_tx_empty());

		bfcr = read8(QSPI_SPBFCR);
		bfcr &= ~QSPI_SPBFCR_RXTRG_MASK;

		n = min(remain, QSPI_BUFFER_SIZE);
		if (n >= QSPI_BUFFER_SIZE) {
			/* sets triggering number to 32 bytes */
			bfcr |= QSPI_SPBFCR_RXTRG_32B;
		} else {
			/* sets triggering number to 1 byte */
			bfcr |= QSPI_SPBFCR_RXTRG_1B;
		}

		for (i = 0; i < n; i++) {
			/* dummy write to generate clock */
			write8(QSPI_SPDR, 0x00);
		}

		if (n >= QSPI_BUFFER_SIZE) {
			/* Wait receive */
			while (!qspi_is_rx_full());
			for (i = 0; i < n; i++)
				*buf++ = read8(QSPI_SPDR);
		} else {
			for (i = 0; i < n; i++) {
				/* Wait receive */
				while (!qspi_is_rx_full());

				*buf++ = read8(QSPI_SPDR);
			}
		}
		remain -= n;
	}
}
