#include "board.h"
#include "scif.h"
#include "io.h"
#include "types.h"

#define SCIFB_BASE	(SCIFB1_BASE_ADDR)
#define SCIFB_BAUD	(115200)
#define SCIFB_CLOCK	(52000000) /* MP clock */

#define SCFSR_ERRORS	(SCIFB_SCBFSR_PER | SCIFB_SCBFSR_FER | \
			 SCIFB_SCBFSR_ER | SCIFB_SCBFSR_BRK)

#define SCFSR_ERROR_CLEAR	(SCIFB_SCBFSR_TEND | SCIFB_SCBFSR_TDFE | \
				 SCIFB_SCBFSR_BRK | SCIFB_SCBFSR_RDF |   \
				 SCIFB_SCBFSR_DR)

static void scifb1_supply_clock(void)
{
	u32 sr, cr;

	sr = read32(CPG_MSTPSR2);
	if (sr & CPG_MSTP2_SCIFB1) {
		cr = read32(CPG_SMSTPCR2);
		write32(CPG_SMSTPCR2, cr & ~CPG_MSTP2_SCIFB1);
	}
}

static void scifb1_pinmux(void)
{
	u32 reg;

	reg = read32(PFC_IPSR16);
	reg &= ~0x0000003f;
	reg |=  0x00000009;
	/* SCIFB1_RXD, SCIFB1_TXD */
	write32(PFC_PMMR, ~reg);
	write32(PFC_IPSR16, reg);

	reg = read32(PFC_GPSR7);
	reg |=  0x00000060;
	/* GP7[6:5] -> Peripheral(SCIFB1) */
	write32(PFC_PMMR, ~reg);
	write32(PFC_GPSR7, reg);

	reg = read32(PFC_PUPR7);
	reg |=  0x00002000;
	/* HRX1 pin pull up */
	write32(PFC_PUPR7, reg);
}

void scif_setup(void)
{
	scifb1_supply_clock();
	scifb1_pinmux();

	write16(SCIFB_BASE + SCIFB_SCBSCR, 0x0030);
	write16(SCIFB_BASE + SCIFB_SCBSMR, 0);

	/* reset FIFO */
	write16(SCIFB_BASE + SCIFB_SCBFCR, 0x0006);
	write16(SCIFB_BASE + SCIFB_SCBFCR, 0);

	/* set baudrate 115200 */
	write8(SCIFB_BASE + SCIFB_SCBBRR, 0x1b);
}

static void scifb_handle_error(void)
{
	read16(SCIFB_BASE + SCIFB_SCBFSR);
	write16(SCIFB_BASE + SCIFB_SCBFSR, SCFSR_ERROR_CLEAR);
}

int scif_getc(void)
{
	u16 fsr;
	int ch;

	do {
		fsr = read16(SCIFB_BASE + SCIFB_SCBFSR);
		if (fsr & SCFSR_ERRORS)
			scifb_handle_error();
	} while (!(fsr & (SCIFB_SCBFSR_DR | SCIFB_SCBFSR_RDF)));


	ch = read16(SCIFB_BASE + SCIFB_SCBFRDR);
	fsr = read16(SCIFB_BASE + SCIFB_SCBFSR);

	write16(SCIFB_BASE + SCIFB_SCBFSR,
		~(SCIFB_SCBFSR_RDF | SCIFB_SCBFSR_DR));

	if (fsr & SCFSR_ERRORS)
		scifb_handle_error();

	return ch;
}

int scif_putc(u8 c)
{
	/* waiting for transmission end */
	while(1) {
		if (read16(SCIFB_BASE + SCIFB_SCBFSR) & SCIFB_SCBFSR_TEND)
			break;
	}

	write8(SCIFB_BASE + SCIFB_SCBFTDR, c);
	write16(SCIFB_BASE + SCIFB_SCBFSR,
		read16(SCIFB_BASE + SCIFB_SCBFSR) & ~SCIFB_SCBFSR_TEND);

	return 0;
}

int scif_puts(const u8 *s)
{
	while (*s) {
		scif_putc(*s);
		s++;
	}

	return 0;
}
